@@ -14,10 +14,17 @@ from .constants import (
     DEFAULT_TIMEOUT,
     FEATURE_CODE,
     MAX_RETRIES,
+    REQUEST_HEADER,
     DefenseModeType,
     DeviceCatagories,
 )
-from .exceptions import EzvizAuthTokenExpired, HTTPError, InvalidURL, PyEzvizError
+from .exceptions import (
+    EzvizAuthTokenExpired,
+    EzvizAuthVerificationCode,
+    HTTPError,
+    InvalidURL,
+    PyEzvizError,
+)
 
 API_ENDPOINT_CLOUDDEVICES = "/api/cloud/v2/cloudDevices/getAll"
 API_ENDPOINT_PAGELIST = "/v3/userdevices/v1/resources/pagelist"
@@ -35,6 +42,9 @@ API_ENDPOINT_SWITCH_DEFENCE_MODE = "/v3/userdevices/v1/group/switchDefenceMode"
 API_ENDPOINT_SWITCH_SOUND_ALARM = "/sendAlarm"
 API_ENDPOINT_SERVER_INFO = "/v3/configurations/system/info"
 API_ENDPOINT_LOGOUT = "/v3/users/logout/v2"
+API_ENDPOINT_PANORAMIC_DEVICES_OPERATION = "/v3/panoramicDevices/operation"
+API_ENDPOINT_UPGRADE_DEVICE = "/v3/upgrades/v1/devices/"
+API_ENDPOINT_SEND_CODE = "/v3/sms/nologin/checkcode"
 
 
 class EzvizClient:
@@ -50,10 +60,12 @@ class EzvizClient:
     ) -> None:
         """Initialize the client object."""
         self.account = account
-        self.password = password
-        self._session = requests.session()
-        # Set Android generic user agent.
-        self._session.headers.update({"User-Agent": "okhttp/3.12.1"})
+        self.password = (
+            hashlib.md5(password.encode("utf-8")).hexdigest() if password else None
+        )  # Ezviz API sends md5 of password
+        self._session = None
+        self.close_session()
+        self._session.headers["sessionId"] = token.get("session_id") if token else None
         self._token = token or {
             "session_id": None,
             "rf_session_id": None,
@@ -63,36 +75,26 @@ class EzvizClient:
         self._timeout = timeout
         self._cameras: dict[str, Any] = {}
 
-    def _login(self, account: str, password: str) -> dict[Any, Any]:
+    def _login(self, smscode: int | None = None) -> dict[Any, Any]:
         """Login to Ezviz API."""
 
         # Region code to url.
         if len(self._token["api_url"].split(".")) == 1:
             self._token["api_url"] = "apii" + self._token["api_url"] + ".ezvizlife.com"
 
-        # Ezviz API sends md5 of password
-        temp = hashlib.md5()
-        temp.update(password.encode("utf-8"))
-        md5pass = temp.hexdigest()
         payload = {
-            "account": account,
-            "password": md5pass,
+            "account": self.account,
+            "password": self.password,
             "featureCode": FEATURE_CODE,
             "msgType": "0",
             "cuName": "SFRDIDEw",
+            "smsCode": smscode,
         }
 
         try:
             req = self._session.post(
                 "https://" + self._token["api_url"] + API_ENDPOINT_LOGIN,
                 allow_redirects=False,
-                headers={
-                    "clientType": "3",
-                    "customno": "1000001",
-                    "clientNo": "web_site",
-                    "appId": "ys7",
-                    "lang": "en",
-                },
                 data=payload,
                 timeout=self._timeout,
             )
@@ -116,6 +118,21 @@ class EzvizClient:
                 + str(req.text)
             ) from err
 
+        if json_result["meta"]["code"] == 200:
+            self._session.headers["sessionId"] = json_result["loginSession"][
+                "sessionId"
+            ]
+            self._token = {
+                "session_id": str(json_result["loginSession"]["sessionId"]),
+                "rf_session_id": str(json_result["loginSession"]["rfSessionId"]),
+                "username": str(json_result["loginUser"]["username"]),
+                "api_url": str(json_result["loginArea"]["apiDomain"]),
+            }
+
+            self._token["service_urls"] = self.get_service_urls()
+
+            return self._token
+
         if json_result["meta"]["code"] == 1100:
             self._token["api_url"] = json_result["loginArea"]["apiDomain"]
             print("Region incorrect!")
@@ -132,42 +149,66 @@ class EzvizClient:
         if json_result["meta"]["code"] == 1015:
             raise PyEzvizError("The user is locked.")
 
-        self._token["session_id"] = str(json_result["loginSession"]["sessionId"])
-        self._token["rf_session_id"] = str(json_result["loginSession"]["rfSessionId"])
-        self._token["username"] = str(json_result["loginUser"]["username"])
-        self._token["api_url"] = str(json_result["loginArea"]["apiDomain"])
-        if not self._token["session_id"]:
-            raise PyEzvizError(
-                f"Login error: Please check your username/password: {req.text}"
+        if json_result["meta"]["code"] == 6002:
+            self.send_mfa_code()
+            raise EzvizAuthVerificationCode(
+                "MFA enabled on account. Please retry with code."
             )
 
-        self._token["service_urls"] = self.get_service_urls()
-
-        return self._token
-
-    def get_service_urls(self) -> Any:
-        """Get Ezviz service urls."""
+        raise PyEzvizError(f"Login error: {json_result['meta']}")
 
+    def send_mfa_code(self) -> bool:
+        """Send verification code."""
         try:
-            req = self._session.get(
-                f"https://{self._token['api_url']}{API_ENDPOINT_SERVER_INFO}",
-                headers={
-                    "sessionId": self._token["session_id"],
-                    "featureCode": FEATURE_CODE,
+            req = self._session.post(
+                "https://" + self._token["api_url"] + API_ENDPOINT_SEND_CODE,
+                data={
+                    "from": self.account,
+                    "bizType": "TERMINAL_BIND",
                 },
                 timeout=self._timeout,
             )
 
             req.raise_for_status()
 
+        except requests.HTTPError as err:
+            raise HTTPError from err
+
+        try:
+            json_output = req.json()
+
+        except ValueError as err:
+            raise PyEzvizError(
+                "Impossible to decode response: "
+                + str(err)
+                + "\nResponse was: "
+                + str(req.text)
+            ) from err
+
+        if json_output.get("meta").get("code") != 200:
+            raise PyEzvizError(
+                f"Could not request MFA code: Got {req.status_code} : {req.text})"
+            )
+
+        return True
+
+    def get_service_urls(self) -> Any:
+        """Get Ezviz service urls."""
+
+        if not self._token["session_id"]:
+            raise PyEzvizError("No Login token present!")
+
+        try:
+            req = self._session.get(
+                f"https://{self._token['api_url']}{API_ENDPOINT_SERVER_INFO}",
+                timeout=self._timeout,
+            )
+            req.raise_for_status()
+
         except requests.ConnectionError as err:
             raise InvalidURL("A Invalid URL or Proxy error occured") from err
 
         except requests.HTTPError as err:
-            if err.response.status_code == 401:
-                # session is wrong, need to relogin
-                self.login()
-
             raise HTTPError from err
 
         if not req.text:
@@ -185,7 +226,7 @@ class EzvizClient:
             ) from err
 
         if json_output.get("meta").get("code") != 200:
-            logging.info("Json request error")
+            raise PyEzvizError(f"Error getting Service URLs: {json_output['meta']}")
 
         service_urls = json_output["systemConfigInfo"]
         service_urls["sysConf"] = service_urls["sysConf"].split("|")
@@ -206,7 +247,6 @@ class EzvizClient:
         try:
             req = self._session.get(
                 "https://" + self._token["api_url"] + API_ENDPOINT_PAGELIST,
-                headers={"sessionId": self._token["session_id"]},
                 params={"filter": page_filter},
                 timeout=self._timeout,
             )
@@ -274,7 +314,6 @@ class EzvizClient:
         try:
             req = self._session.get(
                 "https://" + self._token["api_url"] + API_ENDPOINT_ALARMINFO_GET,
-                headers={"sessionId": self._token["session_id"]},
                 params=params,
                 timeout=self._timeout,
             )
@@ -321,7 +360,6 @@ class EzvizClient:
                 + "/1/1/"
                 + str(status_type)
                 + API_ENDPOINT_SWITCH_STATUS,
-                headers={"sessionId": self._token["session_id"]},
                 data={
                     "enable": enable,
                     "serial": serial,
@@ -359,6 +397,49 @@ class EzvizClient:
 
         return True
 
+    def upgrade_device(self, serial: str, max_retries: int = 0) -> bool:
+        """Upgrade device firmware."""
+        if max_retries > MAX_RETRIES:
+            raise PyEzvizError("Can't gather proper data. Max retries exceeded.")
+
+        try:
+            req = self._session.put(
+                "https://"
+                + self._token["api_url"]
+                + API_ENDPOINT_UPGRADE_DEVICE
+                + serial
+                + "/0/upgrade",
+                timeout=self._timeout,
+            )
+
+            req.raise_for_status()
+
+        except requests.HTTPError as err:
+            if err.response.status_code == 401:
+                # session is wrong, need to relogin
+                self.login()
+                return self.upgrade_device(serial, max_retries + 1)
+
+            raise HTTPError from err
+
+        try:
+            json_output = req.json()
+
+        except ValueError as err:
+            raise PyEzvizError(
+                "Impossible to decode response: "
+                + str(err)
+                + "\nResponse was: "
+                + str(req.text)
+            ) from err
+
+        if json_output.get("meta").get("code") != 200:
+            raise PyEzvizError(
+                f"Could not initiate firmare upgrade: Got {req.status_code} : {req.text})"
+            )
+
+        return True
+
     def sound_alarm(self, serial: str, enable: int = 1, max_retries: int = 0) -> bool:
         """Sound alarm on a device."""
         if max_retries > MAX_RETRIES:
@@ -372,7 +453,6 @@ class EzvizClient:
                 + serial
                 + "/0"
                 + API_ENDPOINT_SWITCH_SOUND_ALARM,
-                headers={"sessionId": self._token["session_id"]},
                 data={
                     "enable": enable,
                 },
@@ -439,41 +519,43 @@ class EzvizClient:
         """Load all devices and build dict per device serial."""
 
         devices = self._get_page_list()
-        result: dict[Any, Any] = {}
+        result: dict[str, Any] = {}
         _res_id = "NONE"
 
         for device in devices["deviceInfos"]:
             _serial = device["deviceSerial"]
-            _res_id = [
+            _res_id_list = {
                 item
-                for item in devices["CLOUD"]
+                for item in devices.get("CLOUD", {}).keys()
                 if devices["CLOUD"][item].get("deviceSerial") == _serial
-            ][0]
+            }
+            _res_id = _res_id_list.pop() if len(_res_id_list) else "NONE"
+
             result[_serial] = {
-                "CLOUD": {_res_id: devices.get("CLOUD", {}).get(_res_id)},
-                "VTM": {_res_id: devices.get("VTM", {}).get(_res_id)},
-                "P2P": devices.get("P2P", {}).get(_serial),
-                "CONNECTION": devices.get("CONNECTION", {}).get(_serial),
-                "KMS": devices.get("KMS", {}).get(_serial),
-                "STATUS": devices.get("STATUS", {}).get(_serial),
-                "TIME_PLAN": devices.get("TIME_PLAN", {}).get(_serial),
-                "CHANNEL": {_res_id: devices.get("CHANNEL", {}).get(_res_id)},
-                "QOS": devices.get("QOS", {}).get(_serial),
-                "NODISTURB": devices.get("NODISTURB", {}).get(_serial),
-                "FEATURE": devices.get("FEATURE", {}).get(_serial),
-                "UPGRADE": devices.get("UPGRADE", {}).get(_serial),
-                "FEATURE_INFO": devices.get("FEATURE_INFO", {}).get(_serial),
-                "SWITCH": devices.get("SWITCH", {}).get(_serial),
-                "CUSTOM_TAG": devices.get("CUSTOM_TAG", {}).get(_serial),
+                "CLOUD": {_res_id: devices.get("CLOUD", {}).get(_res_id, {})},
+                "VTM": {_res_id: devices.get("VTM", {}).get(_res_id, {})},
+                "P2P": devices.get("P2P", {}).get(_serial, {}),
+                "CONNECTION": devices.get("CONNECTION", {}).get(_serial, {}),
+                "KMS": devices.get("KMS", {}).get(_serial, {}),
+                "STATUS": devices.get("STATUS", {}).get(_serial, {}),
+                "TIME_PLAN": devices.get("TIME_PLAN", {}).get(_serial, {}),
+                "CHANNEL": {_res_id: devices.get("CHANNEL", {}).get(_res_id, {})},
+                "QOS": devices.get("QOS", {}).get(_serial, {}),
+                "NODISTURB": devices.get("NODISTURB", {}).get(_serial, {}),
+                "FEATURE": devices.get("FEATURE", {}).get(_serial, {}),
+                "UPGRADE": devices.get("UPGRADE", {}).get(_serial, {}),
+                "FEATURE_INFO": devices.get("FEATURE_INFO", {}).get(_serial, {}),
+                "SWITCH": devices.get("SWITCH", {}).get(_serial, {}),
+                "CUSTOM_TAG": devices.get("CUSTOM_TAG", {}).get(_serial, {}),
                 "VIDEO_QUALITY": {
-                    _res_id: devices.get("VIDEO_QUALITY", {}).get(_res_id)
+                    _res_id: devices.get("VIDEO_QUALITY", {}).get(_res_id, {})
                 },
                 "resourceInfos": [
                     item
-                    for item in devices["resourceInfos"]
-                    if item["deviceSerial"] == _serial
+                    for item in devices.get("resourceInfos")
+                    if item.get("deviceSerial") == _serial
                 ],  # Could be more than one
-                "WIFI": devices["WIFI"].get(_serial),
+                "WIFI": devices.get("WIFI", {}).get(_serial, {}),
                 "deviceInfos": device,
             }
 
@@ -506,7 +588,6 @@ class EzvizClient:
                     "uuid": str(uuid4()),
                     "serial": serial,
                 },
-                headers={"sessionId": self._token["session_id"], "clientType": "1"},
                 timeout=self._timeout,
             )
 
@@ -517,7 +598,41 @@ class EzvizClient:
 
         return req.text
 
-    def login(self) -> dict[Any, Any]:
+    def ptz_control_coordinates(
+        self, serial: str, x_axis: float, y_axis: float
+    ) -> bool:
+        """PTZ Coordinate Move."""
+        if 0 < x_axis > 1:
+            raise PyEzvizError(
+                f"Invalid X coordinate: {x_axis}: Should be between 0 and 1 inclusive"
+            )
+
+        if 0 < y_axis > 1:
+            raise PyEzvizError(
+                f"Invalid Y coordinate: {y_axis}: Should be between 0 and 1 inclusive"
+            )
+
+        try:
+            req = self._session.post(
+                "https://"
+                + self._token["api_url"]
+                + API_ENDPOINT_PANORAMIC_DEVICES_OPERATION,
+                data={
+                    "x": f"{x_axis:.6f}",
+                    "y": f"{y_axis:.6f}",
+                    "deviceSerial": serial,
+                },
+                timeout=self._timeout,
+            )
+
+            req.raise_for_status()
+
+        except requests.HTTPError as err:
+            raise HTTPError from err
+
+        return True
+
+    def login(self, sms_code: int | None = None) -> dict[Any, Any]:
         """Get or refresh ezviz login token."""
         if self._token["session_id"] and self._token["rf_session_id"]:
             try:
@@ -529,7 +644,6 @@ class EzvizClient:
                         "refreshSessionId": self._token["rf_session_id"],
                         "featureCode": FEATURE_CODE,
                     },
-                    headers={"sessionId": self._token["session_id"]},
                     timeout=self._timeout,
                 )
                 req.raise_for_status()
@@ -548,6 +662,18 @@ class EzvizClient:
                     + str(req.text)
                 ) from err
 
+            if json_result["meta"].get("code") == 200:
+
+                self._token["session_id"] = str(json_result["sessionInfo"]["sessionId"])
+                self._token["rf_session_id"] = str(
+                    json_result["sessionInfo"]["refreshSessionId"]
+                )
+
+                if not self._token.get("service_urls"):
+                    self._token["service_urls"] = self.get_service_urls()
+
+                return self._token
+
             if json_result["meta"].get("code") == 403:
                 if self.account and self.password:
                     self._token = {
@@ -558,22 +684,14 @@ class EzvizClient:
                     }
                     return self.login()
 
-                raise EzvizAuthTokenExpired(f"Relogin required: {req.text}")
+                raise EzvizAuthTokenExpired(
+                    f"Token expired, Login with username and password required: {req.text}"
+                )
 
-            self._token["session_id"] = str(json_result["sessionInfo"]["sessionId"])
-            self._token["rf_session_id"] = str(
-                json_result["sessionInfo"]["refreshSessionId"]
-            )
-            if not self._token["session_id"]:
-                raise PyEzvizError(f"Relogin required: {req.text}")
-
-            if not self._token.get("service_urls"):
-                self._token["service_urls"] = self.get_service_urls()
-
-            return self._token
+            raise PyEzvizError(f"Error renewing login token: {json_result['meta']}")
 
         if self.account and self.password:
-            return self._login(account=self.account, password=self.password)
+            return self._login(sms_code)
 
         raise PyEzvizError("Login with account and password required")
 
@@ -582,18 +700,14 @@ class EzvizClient:
         try:
             req = self._session.delete(
                 "https://" + self._token["api_url"] + API_ENDPOINT_LOGOUT,
-                headers={
-                    "featureCode": FEATURE_CODE,
-                    "customno": "1000001",
-                    "sessionId": self._token["session_id"],
-                    "clientNo": "web_site",
-                    "appId": "ys7",
-                },
                 timeout=self._timeout,
             )
             req.raise_for_status()
 
         except requests.HTTPError as err:
+            if err.response.status_code == 401:
+                print("Token is no longer valid. Already logged out?")
+                return True
             raise HTTPError from err
 
         try:
@@ -637,7 +751,6 @@ class EzvizClient:
         try:
             req = self._session.post(
                 "https://" + self._token["api_url"] + API_ENDPOINT_SET_DEFENCE_SCHEDULE,
-                headers={"sessionId": self._token["session_id"]},
                 data={
                     "devTimingPlan": schedulestring,
                 },
@@ -682,7 +795,6 @@ class EzvizClient:
         try:
             req = self._session.post(
                 "https://" + self._token["api_url"] + API_ENDPOINT_SWITCH_DEFENCE_MODE,
-                headers={"sessionId": self._token["session_id"]},
                 data={
                     "groupId": -1,
                     "mode": mode,
@@ -739,7 +851,6 @@ class EzvizClient:
                 "https://"
                 + self._token["api_url"]
                 + API_ENDPOINT_DETECTION_SENSIBILITY,
-                headers={"sessionId": self._token["session_id"]},
                 data={
                     "subSerial": serial,
                     "type": type_value,
@@ -784,7 +895,6 @@ class EzvizClient:
                 "https://"
                 + self._token["api_url"]
                 + API_ENDPOINT_DETECTION_SENSIBILITY_GET,
-                headers={"sessionId": self._token["session_id"]},
                 data={
                     "subSerial": serial,
                 },
@@ -839,7 +949,6 @@ class EzvizClient:
                 + API_ENDPOINT_DEVICES
                 + serial
                 + API_ENDPOINT_ALARM_SOUND,
-                headers={"sessionId": self._token["session_id"]},
                 data={
                     "enable": enable,
                     "soundType": sound_type,
@@ -924,6 +1033,4 @@ class EzvizClient:
             self._session.close()
 
         self._session = requests.session()
-        self._session.headers.update(
-            {"User-Agent": "okhttp/3.12.1"}
-        )  # Android generic user agent.
+        self._session.headers.update(REQUEST_HEADER)  # Reset session.
